unit McTypes;

interface

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls;

type
	TThreadViewColumnType = (
		ctNone,					{ Ȃ }
		ctIcon,					{ ACR }
		ctPriority,			{ dvx }
		ctDepth,				{ [x }
		ctTitle,				{ Xbh^Cg }
		ctRealResCount,	{ X }
		ctResCount,			{ X }
		ctNewResCount,	{ VX }
		ctDateTime,			{ ŏI擾 }
		ctWriteTime,		{ ŏI }
		ctBoardName,		{ {[h }
		ctMemo,					{  }
		ctThreadNum,		{ Xbhԍ }
		ctSelected			{ I }
	);
	TThreadViewSortDirection = (sdAsc{}, sdDec{~});
	TThreadViewColumnTypeName = record
		ColumnType: TThreadViewColumnType;
		ID: string;
		Name: string;
		Caption: string;
		Width: Integer;
		Align: TAlignment;
	end;
	TThreadViewColumnItem = record
    ColumnType: TThreadViewColumnType;
		Width: Integer;
		Caption: string;
		Align: TAlignment;
	end;
	PThreadViewColumnItem = ^TThreadViewColumnItem;
	TThreadViewColumn = class
	private
		FInternalList: TStringList;
    function GetItem(Index: Integer): PThreadViewColumnItem;
    function GetCount: Integer;
    function Get(Index: Integer): string;
    procedure Put(Index: Integer; const Value: string);
	public
		constructor Create;
		destructor Destroy; override;
		function Add(const S: string): Integer;
		procedure Clear;
    function IndexOf(const S: string): Integer;
		property Items[Index: Integer]: PThreadViewColumnItem read GetItem;
		property Count: Integer read GetCount;
		property Strings[Index: Integer]: string read Get write Put; default;
	end;
	TBrowserViewTabType = (ttCurrentTab, ttNewTab);

implementation

const
	ColumnTypeName: array [0 .. 13] of TThreadViewColumnTypeName =
		((ColumnType: ctNone; ID: 'none'; Name: 'Ȃ'; Caption: ''; Width: 0; Align: taLeftJustify),
		 (ColumnType: ctIcon; ID: 'icon'; Name: 'ACR'; Caption: ''; Width: 20; Align: taLeftJustify),
		 (ColumnType: ctPriority; ID: 'priority'; Name: 'dvx'; Caption: 'dvx'; Width: ColumnHeaderWidth; Align: taRightJustify),
		 (ColumnType: ctDepth; ID: 'depth'; Name: '[x'; Caption: 'No.'; Width: 30; Align: taRightJustify),
		 (ColumnType: ctTitle; ID: 'title'; Name: '^Cg'; Caption: '^Cg'; Width: 120; Align: taLeftJustify),
		 (ColumnType: ctRealResCount; ID: 'realrescount'; Name: 'X'; Caption: 'X'; Width: ColumnHeaderWidth; Align: taRightJustify),
		 (ColumnType: ctResCount; ID: 'rescount'; Name: 'X'; Caption: ''; Width: ColumnHeaderWidth; Align: taRightJustify),
		 (ColumnType: ctNewResCount; ID: 'newrescount'; Name: 'VX'; Caption: 'V'; Width: ColumnHeaderWidth; Align: taRightJustify),
		 (ColumnType: ctDateTime; ID: 'datetime'; Name: 'ŏI擾'; Caption: 'ŏI擾'; Width: ColumnHeaderWidth; Align: taLeftJustify),
		 (ColumnType: ctWriteTime; ID: 'writetime'; Name: 'ŏI'; Caption: 'ŏI'; Width: ColumnHeaderWidth; Align: taLeftJustify),
		 (ColumnType: ctBoardName; ID: 'boardname'; Name: '{[h'; Caption: '{[h'; Width: 50; Align: taLeftJustify),
		 (ColumnType: ctMemo; ID: 'memo'; Name: ''; Caption: ''; Width: 50; Align: taLeftJustify),
		 (ColumnType: ctThreadNum; ID: 'threadnum'; Name: 'Xbhԍ'; Caption: 'Xbh'; Width: 50; Align: taRightJustify),
		 (ColumnType: ctSelected; ID: 'select'; Name: 'I'; Caption: 'I'; Width: ColumnHeaderWidth; Align: taLeftJustify));

{ TThreadViewColumn }

function TThreadViewColumn.Add(const S: string): Integer;
var
	Item: PThreadViewColumnItem;
	I: Integer;
	Found: Boolean;
begin
	Found := False;
	New(Item);
	for I := 0 to 13 do
		if ColumnTypeName[I].ID = LowerCase(S) then
		begin
			Item.ColumnType := ColumnTypeName[I].ColumnType;
			Item.Width := ColumnTypeName[I].Width;
			Item.Align := ColumnTypeName[I].Align;
			Item.Caption := ColumnTypeName[I].Caption;
			Found := True;
			Break;
		end;

	if not Found then
	begin
		Dispose(Item);
		Result := -1;
		Exit;
	end;
	
	Result := FInternalList.AddObject(S, TObject(Item));
end;

procedure TThreadViewColumn.Clear;
var
	I: Integer;
begin
	for I := 0 to FInternalList.Count - 1 do
		Dispose(PThreadViewColumnItem(FInternalList.Objects[I]));

	FInternalList.Clear;
end;

constructor TThreadViewColumn.Create;
begin
	inherited;
	FInternalList := TStringList.Create;
end;

destructor TThreadViewColumn.Destroy;
begin
	FInternalList.Free;
  inherited;
end;

function TThreadViewColumn.Get(Index: Integer): string;
begin
	Result := FInternalList.Strings[Index];
end;

function TThreadViewColumn.GetCount: Integer;
begin
	Result := FInternalList.Count;
end;

function TThreadViewColumn.GetItem(Index: Integer): PThreadViewColumnItem;
begin
	Result := PThreadViewColumnItem(FInternalList.Objects[Index]);
end;

function TThreadViewColumn.IndexOf(const S: string): Integer;
begin
	Result := FInternalList.IndexOf(LowerCase(S));
end;

procedure TThreadViewColumn.Put(Index: Integer; const Value: string);
begin
	FInternalList.Strings[Index] := Value;
end;

end.

