unit ThreadView;

interface

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
	ComCtrls, McConfig, McTypes, BBSUtils, IdAsyncHTTP, StdCtrls, Buttons,
  ImgList, Variants, SyncObjs;

type
	EThreadViewError = class(Exception);
	TThreadViewLoadThread = class;
	TThreadViewFrame = class(TFrame)
    ListView: TListView;
    TabImageList: TImageList;
    procedure ListViewCustomDrawItem(Sender: TCustomListView;
      Item: TListItem; State: TCustomDrawState; var DefaultDraw: Boolean);
    procedure ListViewColumnClick(Sender: TObject; Column: TListColumn);
    procedure ListViewCompare(Sender: TObject; Item1, Item2: TListItem;
      Data: Integer; var Compare: Integer);
    procedure ListViewDblClick(Sender: TObject);
  private
		{ Private 錾 }
		FColumn: TThreadViewColumn;
		FConfig: TMaCherieConfig;
		FLoadThread: TThreadViewLoadThread;
		FIndexList: T2chIndexList;
		FStorageList: T2chIndexList;
		FSubject: T2chSubject;
		FSubjectIniFile: T2chSubjectIniFile;
		FBeforeSubject: T2chSubject;
		FSelectedBoard: string;
		FServerName: string;
		FBoardPath: string;
		FBoardName: string;
		FSortColumn: TThreadViewColumnType;
		FSortDirection: TThreadViewSortDirection;
		FListLock: TCriticalSection;
    function GetColumnString: string;
		procedure SetColumnString(const Value: string);
		procedure RefreshColumn;
		procedure SetConfig(const Value: TMaCherieConfig);
		procedure AddOneLine(const ALineText: string);
		procedure SetSelectedBoard(const Value: string);
    function GetSelectedThread: T2chThreadIniFile;
		procedure ThreadCatchException(Message: string);
		procedure EndSubjectLoad;
	public
		{ Public 錾 }
    constructor Create(AOwner: TComponent); override;
		destructor Destroy; override;
		procedure RefreshListItem(Item: TListItem; Thread: T2chThreadIniFile);
		property ColumnString: string read GetColumnString write SetColumnString;
		property Config: TMaCherieConfig read FConfig write SetConfig;
		property SelectedBoard: string read FSelectedBoard write SetSelectedBoard;
		property BoardName: string read FBoardName write FBoardName;
		property SelectedThread: T2chThreadIniFile read GetSelectedThread;
	end;
	TThreadViewLoadThread = class(TThread)
	private
		FOwner: TThreadViewFrame;
		FHTTP: TIdAsyncHTTP;
		FLine: string;
    FURL: string;
		FDisconnect: Boolean;
		FExceptionMessage: string;
		FConfig: TMaCherieConfig;
    FLastModified: string;
    FETag: string;
		FNotModified: Boolean;
		procedure SubjectOnLine(ASender: TObject; const ALineText, ATerminator: string; Param: Pointer);
		procedure OnLine;
		procedure CatchException;
		procedure EndSubjectLoad;
	protected
		procedure Execute; override;
	public
		constructor Create(AOwner: TThreadViewFrame);
		destructor Destroy; override;
		property Config: TMaCherieConfig read FConfig write FConfig;
		property URL: string read FURL write FURL;
		property Disconnect: Boolean read FDisconnect write FDisconnect;
		property LastModified: string read FLastModified write FLastModified;
		property ETag: string read FETag write FETag;
		property NotModified: Boolean read FNotModified write FNotModified;
	end;

implementation

uses AwkFunc, IdTCPConnection, IdGlobal, McMain;

{$R *.dfm}

{ TThreadViewFrame }

procedure TThreadViewFrame.AddOneLine(const ALineText: string);
var
	I: Integer;
	Thread: T2chThreadIniFile;
	Item: TListItem;
	LogPath: string;
begin
	FListLock.Enter;
	try
		I := FSubject.Add(ALineText);

		LogPath := Format('%s\%s\%s.ini',
			[URLToPath(FServerName), FBoardPath, FormatCurr('000000000', FSubject[I].ThreadNum)]);

		Thread := FStorageList.LogPathOfObject(LogPath);
		if Thread <> nil then
		begin
			Item := TListItem(Thread.ListItem);
			ListView.Items.Delete(Item.Index);
			Thread.Depth := FSubject[I].Depth;
			Thread.RealResCount := FSubject[I].ResCount;
		end else
		begin
			Thread := T2chThreadIniFile.Create;
			Thread.ThreadTitle := HTML2String(FSubject[I].Title);
			Thread.RealResCount := FSubject[I].ResCount;
			Thread.Depth := FSubject[I].Depth;
			Thread.ThreadNum := FSubject[I].ThreadNum;
			Thread.BoardName := FBoardName;
			Thread.ServerName := FServerName;
			Thread.BoardPath := FBoardPath;
			Thread.ResCount := 0;

			FIndexList.Add(Thread);
		end;

		Thread.ResState := rsNone;

		if Thread.Freezed then
			Thread.ResState := rsFreezed
		else if Thread.ResCount > 0 then
			if Thread.ResCount < Thread.RealResCount then
				Thread.ResState := rsNew	{ V }
			else if Thread.ResCount > Thread.RealResCount then
				Thread.ResState := rsError;	{ G[ }
		if FBeforeSubject.Count > 0 then
			if FBeforeSubject.IndexOf(Thread.ThreadNum) < 0 then
				Thread.ResState := rsNewThread;

		Item := ListView.Items.Insert(FSubject[I].Depth - 1);
	finally
		FListLock.Leave;
	end;
	RefreshListItem(Item, Thread);
end;

constructor TThreadViewFrame.Create(AOwner: TComponent);
begin
	inherited;
	FSubject := T2chSubject.Create;
	FBeforeSubject := T2chSubject.Create;
	FBeforeSubject.Sorted := True;
	FSubjectIniFile := T2chSubjectIniFile.Create;
	FIndexList := T2chIndexList.Create(True);
	FStorageList := T2chIndexList.Create(True);
	FColumn := TThreadViewColumn.Create;
	FLoadThread := TThreadViewLoadThread.Create(Self);
	FListLock := TCriticalSection.Create;
	
	FSortColumn := ctNone;
	FSortDirection := sdAsc;
	ColumnString := 'icon,priority,depth,title,realrescount,rescount,newrescount,datetime,writetime,boardname,memo';
end;

destructor TThreadViewFrame.Destroy;
begin
	FLoadThread.Disconnect := True;
	FLoadThread.Terminate;
	if FLoadThread.Suspended then
		FLoadThread.Resume;
	FLoadThread.WaitFor;
	FLoadThread.Free;
	
	FSubject.Free;
	FSubjectIniFile.Free;
	FBeforeSubject.Free;
	FStorageList.Clear;
	FIndexList.Free;
	FColumn.Free;
	FListLock.Free;
	
  inherited;
end;

function TThreadViewFrame.GetColumnString: string;
var
	I: Integer;
begin
	Result := '';
	for I := 0 to FColumn.Count - 1 do
	begin
		if I > 0 then
			Result := Result + ',';
		Result := Result + FColumn[I];
	end;
end;

procedure TThreadViewFrame.RefreshColumn;
var
	I: Integer;
begin
	FListLock.Enter;
	try
		ListView.Columns.Clear;
		for I := 0 to FColumn.Count - 1 do
		begin
			with ListView.Columns.Add do
			begin
				Caption := FColumn.Items[I].Caption;
				Width := FColumn.Items[I].Width;
				Alignment := FColumn.Items[I].Align;
			end;
		end;
	finally
		FListLock.Leave;
	end;
end;

procedure TThreadViewFrame.RefreshListItem(Item: TListItem;
	Thread: T2chThreadIniFile);
var
	I: Integer;
	Str: string;
begin
	FListLock.Enter;
	try
		Item.SubItems.Clear;
		for I := 0 to FColumn.Count - 1 do
		begin
			Str := '';
			case FColumn.Items[I].ColumnType of
				ctDepth:
					if Thread.Depth >= 0 then
						Str := IntToStr(Thread.Depth);
				ctTitle: Str := Thread.ThreadTitle;
				ctRealResCount: Str := IntToStr(Thread.RealResCount);
				ctBoardName: Str := Thread.BoardName;
			end;

			if Thread.Linked then
			begin
				case FColumn.Items[I].ColumnType of
					ctNewResCount: Str := IntToStr(Thread.NewResCount);
					ctResCount: Str := IntToStr(Thread.ResCount);
					ctMemo: Str := Thread.Memo;
					ctDateTime:
						if Thread.DateTime <> -1 then
							Str := MyDateTimeToStr(Thread.DateTime);
					ctWriteTime:
						if Thread.WriteTime <> -1 then
							Str := MyDateTimeToStr(Thread.WriteTime);
				end;
			end;

			if I = 0
				then Item.Caption := Str
				else Item.SubItems.Add(Str);
		end;

		case Thread.ResState of
			rsNew: Item.StateIndex := 3;
			rsNewThread: Item.StateIndex := 1;
			rsError: Item.StateIndex := 4;
			rsLoading: Item.StateIndex := 0;
			rsFreezed: Item.StateIndex := 5;
			rsNone:
				if Thread.Linked
					then Item.StateIndex := 2
					else Item.StateIndex := -1;
		end;
		Item.Data := Pointer(Thread);
		Thread.ListItem := Pointer(Item);
	finally
		FListLock.Leave;
	end;
end;

procedure TThreadViewFrame.SetColumnString(const Value: string);
var
	List: TStringList;
	I: Integer;
begin
	List := TStringList.Create;
	try
		FColumn.Clear;
		
		if AWKSplit(',', Value, List) < 0 then
			if FColumn.Add(Value) < 0 then
				raise EThreadViewError.Create(Value + ': IDႢ܂')
			else
				Exit;

		for I := 0 to List.Count - 1 do
			if FColumn.Add(List[I]) < 0 then
				raise EThreadViewError.Create(List[I] + ': IDႢ܂');

		RefreshColumn;
	finally
		List.Free;
	end;
end;

procedure TThreadViewFrame.SetConfig(const Value: TMaCherieConfig);
begin
	FConfig := Value;
	FLoadThread.Config := Value;
end;

procedure TThreadViewFrame.SetSelectedBoard(const Value: string);
var
	I: Integer;
	SubjectPath: string;
begin
	{ Xbh̒~ }
	Assert(not FLoadThread.Terminated);
	if FLoadThread.Terminated then Exit;
	FLoadThread.Disconnect := True;
	while not FLoadThread.Suspended do
		Application .ProcessMessages;

	FSelectedBoard := Value;

	FListLock.Enter;
	try
		FIndexList.Clear;
		FSubject.Clear;
		FBeforeSubject.Clear;
		FSubjectIniFile.Clear;
		ListView.Clear;
	finally
		FListLock.Leave;
	end;
	
	{ \[g֌W̏ }
	FSortColumn := ctNone;
	FSortDirection := sdAsc;

	ExtractServerNameAndBoardPath(FSelectedBoard, FServerName, FBoardPath);
	FStorageList.Clear;
	FStorageList.CreateStorageList(FConfig.GetBaseDir + 'logx', FServerName, FBoardPath);

	FConfig.Enter;
	try
		SubjectPath := FConfig.GetBaseDir +
			Format('logx\%s\%s\subject.txt', [URLToPath(FServerName), FBoardPath]);
	finally
		FConfig.Leave;
	end;

	try
		FBeforeSubject.LoadFromFile(SubjectPath);
		FSubjectIniFile.LoadFromFile(ChangeFileExt(SubjectPath, '.ini'));
	except
		on EFOpenError do ;
		else
			raise;
	end;
	
	for I := 0 to FStorageList.Count - 1 do
		RefreshListItem(ListView.Items.Add, FStorageList[I]);

	FLoadThread.ETag := FSubjectIniFile.ETag;
	FLoadThread.LastModified := FSubjectIniFile.LastModified;
	FLoadThread.URL := Format('http://%s/subject.txt', [Value]);
	FLoadThread.Resume;
end;

procedure TThreadViewFrame.ListViewCustomDrawItem(Sender: TCustomListView;
  Item: TListItem; State: TCustomDrawState; var DefaultDraw: Boolean);
var
	Rect: TRect;
	BkColor: TColor;
begin
	Rect := Item.DisplayRect(drBounds);
	Rect.Right := ListView.ClientWidth;

	if (ListView.Items.IndexOf(Item) mod 2) = 0 then
		BkColor := $E7EFEF
	else
		BkColor := clWindow;

	with ListView.Canvas do
	begin
		Brush.Color := clBlack;	{Ȃ[Aʂ̐Fɐݒ肷Ƃ܂}
		Brush.Color := BkColor;
		FillRect(Rect);
	end;
end;

function TThreadViewFrame.GetSelectedThread: T2chThreadIniFile;
begin
	Result := nil;

	if ListView.Selected = nil then Exit;
	if ListView.Selected.Data = nil then Exit;

	Result := T2chThreadIniFile(ListView.Selected.Data);
end;

procedure TThreadViewFrame.ThreadCatchException(Message: string);
begin
	ShowMessage(Message);
end;

procedure TThreadViewFrame.EndSubjectLoad;
var
	SubjectPath: string;
	Subject: T2chSubject;
	I: Integer;
begin
	Subject := nil;

	FConfig.Enter;
	try
		SubjectPath := Config.GetBaseDir + Format('logx\%s\%s', [URLToPath(FServerName), FBoardPath]);
	finally
		FConfig.Leave;
	end;

	{ subject.txtɕύXƂ̏ }
	if FLoadThread.NotModified then
	begin
		{ łɗƂĂsubject.txtǂݍ݁A\ }
		try
			Subject := T2chSubject.Create;
			Subject.LoadFromFile(IncludeTrailingPathDelimiter(SubjectPath) + 'subject.txt');
			for I := 0 to Subject.Count - 1 do
				AddOneLine(Subject.Lines[I]);
		finally
			Subject.Free;
		end;
	end else
	begin
		ForceDirectories(SubjectPath);
		FSubject.SaveToFile(IncludeTrailingPathDelimiter(SubjectPath) + 'subject.txt');
		FSubjectIniFile.ETag := FLoadThread.ETag;
		FSubjectIniFile.LastModified := FLoadThread.LastModified;
		FSubjectIniFile.SaveToFile(IncludeTrailingPathDelimiter(SubjectPath) + 'subject.ini');
	end;
end;

procedure TThreadViewFrame.ListViewColumnClick(Sender: TObject;
	Column: TListColumn);
var
	Shift: Boolean;
	I, Index: Integer;
	ColumnType: TThreadViewColumnType;
begin
	Index := Column.Index;
	Shift := GetAsyncKeyState(VK_SHIFT) <> 0;
	ColumnType := FColumn.Items[Index].ColumnType;

	if (ColumnType = ctTitle) and Shift then
		ColumnType := ctThreadNum
	else if (ColumnType = ctIcon) and Shift then
		ColumnType := ctSelected;

	if FSortColumn = ColumnType then
	begin
		if FSortDirection = sdAsc
			then FSortDirection := sdDec
			else FSortDirection := sdAsc;
	end else
	begin
		FSortColumn := ColumnType;
		FSortDirection := sdAsc;
	end;


	with ListView do
		for I := 0 to Items.Count - 1 do
			T2chThreadIniFile(Items[I].Data).SortIndex := I;

	FListLock.Enter;
	try
		ListView.CustomSort(nil, 0);
	finally
		FListLock.Leave;
	end;
end;

procedure TThreadViewFrame.ListViewDblClick(Sender: TObject);
var
	Thread: T2chThreadIniFile;
begin
	Thread := GetSelectedThread;
	if Thread = nil then Exit;
	MaCherieMainForm.BrowserOpen(ttCurrentTab, Thread);
end;

procedure TThreadViewFrame.ListViewCompare(Sender: TObject; Item1,
	Item2: TListItem; Data: Integer; var Compare: Integer);
var
	Val1, Val2: Variant;
	Thread1, Thread2: T2chThreadIniFile;
{ ResStateToIndex }
	function ResStateToIndex(Thread: T2chThreadIniFile): Integer;
	begin
		Result := 6;
		case Thread.ResState of
			rsNone:
				if Thread.Linked
					then Result := 3
					else Result := 6;
			rsNew: Result := 1;
			rsNewThread: Result := 2;
			rsFreezed: Result := 4;
			rsError: Result := 5;
		end;

		if (Thread.ResState = rsNone) and Thread.Linked then
			Result := 4;
	end;
{ SelectedToIndex }
	function SelectedToIndex(Item: TListItem): Integer;
	begin
		Result := 1;
		if Item.Selected then
			Result := 0;
	end;
{ DepthToIndex }
	function DepthToIndex(Thread: T2chThreadIniFile): Integer;
	begin
		Result := MaxInt;
		if Thread.Depth >= 0 then
			Result := Thread.Depth;
	end;
begin
	Thread1 := T2chThreadIniFile(Item1.Data);
	Thread2 := T2chThreadIniFile(Item2.Data);

	Val1 := Null;
	Val2 := Null;

	if FSortColumn = ctPriority then
	begin
		Val1 := Thread1.Priority;
		Val2 := Thread2.Priority;
	end
	else if FSortColumn = ctDepth then
	begin
		Val1 := DepthToIndex(Thread1);
		Val2 := DepthToIndex(Thread2);
	end
	else if FSortColumn = ctTitle then
	begin
		Val1 := Thread1.ThreadTitle;
		Val2 := Thread2.ThreadTitle;
	end
	else if FSortColumn = ctRealResCount then
	begin
		Val1 := Thread1.RealResCount;
		Val2 := Thread2.RealResCount;
	end
	else if FSortColumn = ctResCount then
	begin
		Val1 := Thread1.ResCount;
		Val2 := Thread2.ResCount;
	end
	else if FSortColumn = ctNewResCount then
	begin
		Val1 := Thread1.NewResCount;
		Val2 := Thread2.NewResCount;
	end
	else if FSortColumn = ctDateTime then
	begin
		Val1 := Thread1.DateTime;
		Val2 := Thread2.DateTime;
	end
	else if FSortColumn = ctWriteTime then
	begin
		Val1 := Thread1.WriteTime;
		Val2 := Thread2.WriteTime;
	end
	else if FSortColumn = ctBoardName then
	begin
		Val1 := Thread1.BoardName;
		Val2 := Thread2.BoardName;
	end
	else if FSortColumn = ctMemo then
	begin
		Val1 := Thread1.Memo;
		Val2 := Thread2.Memo;
	end
	else if FSortColumn = ctThreadNum then
	begin
		Val1 := Thread1.ThreadNum;
		Val2 := Thread2.ThreadNum;
	end
	else if FSortColumn = ctIcon then
	begin
		Val1 := ResStateToIndex(Thread1);
		Val2 := ResStateToIndex(Thread2);
	end
	else if FSortColumn = ctSelected then
	begin
		Val1 := SelectedToIndex(Item1);
		Val2 := SelectedToIndex(Item2);
	end;


	Assert(Val1 <> Null);
	Assert(Val2 <> Null);

	if Val1 = Val2 then
		Compare := 0
	else
		if Val1 < Val2 then
			Compare := -1
		else
			Compare := 1;

	if FSortDirection = sdDec then
		Compare := -Compare;

	if Compare = 0 then
		if Thread1.SortIndex < Thread2.SortIndex then
			Compare := -1
		else
			Compare := 1;
end;

{ TThreadViewLoadThread }

procedure TThreadViewLoadThread.CatchException;
begin
	FOwner.ThreadCatchException(FExceptionMessage);
end;

constructor TThreadViewLoadThread.Create(AOwner: TThreadViewFrame);
begin
	inherited Create(True);
	FOwner := AOwner;
	FHTTP := TIdAsyncHTTP.Create(nil);
end;

destructor TThreadViewLoadThread.Destroy;
begin
	FHTTP.Free;
	inherited;
end;

procedure TThreadViewLoadThread.EndSubjectLoad;
begin
	FOwner.EndSubjectLoad;
end;

procedure TThreadViewLoadThread.Execute;
begin
	while True do
	begin
		if Terminated then Exit;

		FDisconnect := False;
		FNotModified := False;
		
		FConfig.Enter;
		try
			with FHTTP do
			begin
				Request.Clear;
				Request.UserAgent := FConfig.CommParams.UserAgent;
				if FConfig.CommParams.ProxyParams.ProxyServer <> '' then
					ProxyParams.Assign(FConfig.CommParams.ProxyParams)
				else begin
					ProxyParams.ProxyServer := '';
					ProxyParams.ProxyPort := 80;
				end;
				ReadTimeout := FConfig.CommParams.ReadTimeOut;
				Request.AcceptEncoding := 'gzip';
				if FLastModified <> '' then
					Request.LastModified := GMTToLocalDateTime(FLastModified);
			end;
		finally
			FConfig.Leave;
		end;
		
		try
			with FHTTP do
			begin
				Response.Clear;
				Get(FURL, SubjectOnLine);
			end;
		except
			on E: EIdAsyncHTTPProtocolException do
				if E.ReplyErrorCode = 304 then	{ Not Modified }
					FNotModified := True	{ ύXtOTrueɂ }
				else begin
					FExceptionMessage := E.ClassName + ': ' + E.Message;
					Synchronize(CatchException);
				end;
			on E: Exception do
			begin
				FExceptionMessage := E.ClassName + ': ' + E.Message;
				Synchronize(CatchException);
			end;
		end;

		with FHTTP do
		begin
			FETag := Response.RawHeaders.Values['ETag'];
			FLastModified := Response.RawHeaders.Values['Last-Modified'];
		end;

		if not FDisconnect then
			Synchronize(EndSubjectLoad);

		if FDisconnect then
			FDisconnect := False;

		if Terminated then Exit;

		Suspend;
	end;
end;

procedure TThreadViewLoadThread.OnLine;
begin
	FOwner.AddOneLine(FLine);
end;

procedure TThreadViewLoadThread.SubjectOnLine(ASender: TObject;
	const ALineText, ATerminator: string; Param: Pointer);
begin
	FLine := ALineText;
	Synchronize(OnLine);
	if FDisconnect then
		FHTTP.ForceDisconnect;
end;

end.

